{
    Файл является частью библиотеки RubiRobotLib

    © 2017-2021 Слинкин Д.А.

    Класс чтения-записи данных для sysfs-файлов

    Подробная информация о лицензировании находится в файле README,
    входящем в состав библиотеки.

    ДАННОЕ ПРОГРАММНОЕ ОБЕСПЕЧЕНИЕ ПРЕДОСТАВЛЯЕТСЯ «КАК ЕСТЬ», БЕЗ
    КАКИХ-ЛИБО ГАРАНТИЙ, ЯВНО ВЫРАЖЕННЫХ ИЛИ ПОДРАЗУМЕВАЕМЫХ,
    ВКЛЮЧАЯ ГАРАНТИИ ТОВАРНОЙ ПРИГОДНОСТИ, СООТВЕТСТВИЯ ПО ЕГО
    КОНКРЕТНОМУ НАЗНАЧЕНИЮ И ОТСУТСТВИЯ НАРУШЕНИЙ, НО НЕ ОГРАНИЧИВАЯСЬ 
    ИМИ. НИ В КАКОМ СЛУЧАЕ АВТОРЫ ИЛИ ПРАВООБЛАДАТЕЛИ НЕ НЕСУТ
    ОТВЕТСТВЕННОСТИ ПО КАКИМ-ЛИБО ИСКАМ, ЗА УЩЕРБ ИЛИ ПО ИНЫМ ТРЕБОВАНИЯМ,
    В ТОМ ЧИСЛЕ, ПРИ ДЕЙСТВИИ КОНТРАКТА, ДЕЛИКТЕ ИЛИ ИНОЙ СИТУАЦИИ,
    ВОЗНИКШИМ ИЗ-ЗА ИСПОЛЬЗОВАНИЯ ПРОГРАММНОГО ОБЕСПЕЧЕНИЯ ИЛИ ИНЫХ
    ДЕЙСТВИЙ С ПРОГРАММНЫМ ОБЕСПЕЧЕНИЕМ.

 **********************************************************************}

{$I uev3defines.inc}

unit uev3sysfs;

interface
uses uev3,SysUtils;

type
    { TEv3Data }

    TEv3Data=class(TEv3)
      private
       f:longint;
       fowner:TObject;
       Fmode: TEv3DataMode;
       fWriteInterval,fReadInterval:dword;
       fn:string;
       e:longint;
       prevData:string;
       //prevTime,prevGetTime:TDateTime;
       function getiData: longint;
       function getsData: shortstring;
       function getwiData: longint;
       function getwsData: string;
       procedure setiData(AValue: longint);
       procedure Setmode(AValue: TEv3DataMode);
       procedure setsData(AValue: shortstring);
       procedure setwiData(AValue: longint);
       procedure setwsData(AValue: string);
       procedure beepE(port:word; len:word);
       function endGetInterval:boolean;
      public
       skipInterval:boolean;
       // 0.2.5 Последние сохраненнные и загруженные данные
       sStored,sLoaded:string;
       iStored,iLoaded:longint;
       // 0.2.5 Факт как минимум одного сохранения/загрузки
       isStored,isLoaded:boolean;
       function waitEvent(event:string; timeout:dword=high(dword)):boolean; override;
       constructor Create(filename:string; owner:TObject=nil;_mode:TEv3DataMode=ev3ReadOnly);
       procedure setIntervals(R,W:integer);
       procedure Reopen(filename:string);
       destructor destroy;override;
       property mode:TEv3DataMode read Fmode write Setmode;
       property sData:shortstring read getsData write setsData;
       property iData:longint read getiData write setiData;
       property wsData:string read getwsData write setwsData;
       property wiData:longint read getwiData write setwiData;
       procedure hardPutData(data:shortstring);
       // v0.2.5 - переоткрывает файл и обнуляет таймеры
       procedure ReInit();
       // v0.2.5 - пишет-читает строку в файл девайса
       class procedure WriteStrToFile(filename:shortstring; data:shortstring);
       class function ReadStrFromFile(filename:shortstring):shortstring;
    end;

    Ev3Exception=class(Exception)
    end;

    Ev3OpenException=class(Ev3Exception)
    end;

    Ev3ReadException=class(Ev3Exception)
    end;

    Ev3WriteException=class(Ev3Exception)
    end;

implementation
uses
  BaseUnix,Unix,errors,dateutils,uev3Devices
  {$IFDEF ev3SoundUse}
    ,rubiroSound
  {$endif}
  ;

__WARNGLOBALOFF

{ TEv3Data }

function TEv3Data.getiData: longint;
begin
 try
  result:=strtoint(sData);
  iLoaded:=result;
 except
  result:=0;
  iLoaded:=0;
 end;
end;

function TEv3Data.getsData: shortstring;
var //buf:shortstring;
    cnt:integer;
begin
 result:='';
 _lock(9);
 try
   //buf:='';
   cnt:=0;
   while true do begin
    inc(cnt);// writeln('---- ',cnt);
    fpseterrno(0);
//    buf[0]:=chr(fppread(f,buf[1],255,0));
//    dec(buf[0]);
//    result:=buf;
    result[0]:=chr(fppread(f,result[1],255,0));
    dec(result[0]);
    e:=fpgeterrno;
    if (e=0)or(ev3Finalized) then break;
    if (cnt>=countMaxRefreshFactory)or(fowner=nil) then begin
      raise ev3ReadException.create('TEv3Data Read Error: file - '+fn+' ; errorcode - '+inttostr(e)+' ; '+strerror(e));
    end;
    // произошел дисконнект, обновляем фабрику, пока устройство снова не подключится
    if beeperror then beepE(ord((fowner as tev3Device).Port),600 div cnt);
    sleep(errorAttrInterval);
    ev3FactoryRefresh;
   end;
   if not skipInterval then _timerWait(fReadInterval,9);
   _timerStart(9);
//   WD(123);
 finally
   sLoaded:=result;
   isLoaded:=true;
   _unlock(9);
 end;
end;

function TEv3Data.getwiData: longint;
begin
 result:=strtoint(wsData);
end;

function TEv3Data.getwsData: string;
begin
 waitEvent('',high(dword));
 result:=prevdata;
end;

procedure TEv3Data.setiData(AValue: longint);
begin
  iStored:=AValue;
  sdata:=inttostr(Avalue);
end;

procedure TEv3Data.Setmode(AValue: TEv3DataMode);
begin
  if Fmode=AValue then Exit;
  Fmode:=AValue;
  reopen(fn);
end;

procedure TEv3Data.setsData(AValue: shortstring);
 var cnt:integer;
begin
//  writeln('AvaluE - ',Avalue);
 sStored:=AValue;
 isStored:=true;
 AValue+=#10;
 cnt:=0;
 _lock(9);
 try
   while true do begin
    inc(cnt);
    fpseterrno(0);
    fppwrite(f,AValue[1],length(Avalue),0);
    e:=fpgeterrno;
    if (e=0)or(ev3Finalized) then break;
    if (cnt>=countMaxRefreshFactory)or(fowner=nil)then begin
      raise Ev3WriteException.create('TEv3Data Write Error: file - '+fn+' ; errorcode - '+inttostr(e)+' ; '+strerror(e));
    end;
    // произошел дисконнект, обновляем фабрику, пока устройство снова не подключится
    if beeperror then beepE(ord((fowner as tev3Device).Port),600 div cnt);//ev3Sound.beep(900-ord((fowner as tev3Device).Port)*100,600 div cnt);
    sleep(errorAttrInterval);
    ev3FactoryRefresh;
   end;
   if not skipInterval then _timerWait(fWriteInterval,8);
   _timerStart(8);
 finally
   _unlock(9);
 end;
end;

procedure TEv3Data.Reopen(filename: string);
 var m:cint;
begin
  _lock(9);
  try
    fpseterrno(0);
  //  writeln(fn,' + ',fmode,' - ',fpgeterrno );
    if f>0 then begin
  //   writeln('>0!');
     fpclose(f);
    end;
    fpseterrno(0);
    case fmode of
     ev3ReadOnly: m:=O_RdOnly;
     ev3WriteOnly:m:=O_WrOnly;
     ev3ReadWrite:m:=O_RdWr;
    end;
  //  writeln(fn,' - ',fmode,' - ',fpgeterrno );
    fn:=filename;
    f:=fpopen(fn,m);
    e:=fpgeterrno;
    if e<>0 then begin
      sleep(errorInitInterval);
      fpseterrno(0);
      f:=fpopen(fn,m);
      e:=fpgeterrno;
    end;
    if e<>0 then begin
      raise Ev3OpenException.create('TEv3Data Open Error: file - '+fn+' ; mode - '+inttostr(ord(fmode))+' ; errorcode - '+inttostr(e)+' ; '+strerror(e));
    end;
  finally
    _unlock(9);
  end;
end;

procedure TEv3Data.ReInit;
begin
  _timerClear(8);
  _timerClear(9);
  reopen(fn);
end;

procedure TEv3Data.setwiData(AValue: longint);
begin
 iData:=AValue;
end;

procedure TEv3Data.setwsData(AValue: string);
begin
 sData:=AValue;
end;

procedure TEv3Data.beepE(port: word; len:word);
 var tone,cnt:integer;
begin
  if port in [1..4] then begin
   tone:=400; cnt:=port;
  end
  else begin
   tone:=200; cnt:=port div 2;
  end;
{$IFDEF ev3SoundUse}
  case cnt of
   1: ev3sound.beep(tone,len);
   2: ev3sound.beep([tone,len div 2,10,tone,len div 2,10]);
   3: ev3sound.beep([tone,len div 3,10,tone,len div 3,10,tone,len div 3,10]);
   4: ev3sound.beep([tone,len div 4,10,tone,len div 4,10,tone,len div 4,10,tone,len div 4,10]);
  end;
{$endif}
end;

function TEv3Data.endGetInterval: boolean;
// var dt:TDateTime;
begin
 if freadinterval=0 then exit(true);
// dt:=now();
// result:=MilliSecondsBetween(prevGetTime,dt)>fReadInterval;
// if result then prevGetTime:=dt;
// v0.2
 result:=_timer(9)>fReadInterval;
 if result then _timerStart(9);
end;

function TEv3Data.waitEvent(event: string; timeout: dword=high(dword)): boolean;
 var buf:string;
     tm:double;
begin
 //dt:=now();
 // v0.2
 _timerStart(tm);
 event:=event;
 while true do begin
  buf:=sData;
  if buf=prevdata then begin
//    if MilliSecondsBetween(now(),dt)>=timeout then exit(false);
//    sleep(waitInterval);
// v0.2
    if _timer(tm)>=timeout then exit(false);
    sleep(waitInterval);
    continue;
  end;
  prevdata:=buf;
  exit(true);
 end;
end;

constructor TEv3Data.Create(filename: string; owner: TObject;
  _mode: TEv3DataMode);
begin
  inherited Create;
  fmode:=_mode; fowner:=owner;
  setIntervals(getInterval,setInterval);
 // writeln('Create ',filename);
  reopen(filename);
  if f<0 then begin
//    writeln ('FAIL');
    fail;
  end;
end;

procedure TEv3Data.setIntervals(R, W: integer);
begin
 fWriteInterval:=W;
 fReadInterval:=R;
 if (w=0) and (r=0) then skipInterval:=true
 else skipInterval:=false;
end;

destructor TEv3Data.destroy;
begin
  fpclose(f); fpseterrno(0);
  inherited destroy;
end;

procedure TEv3Data.hardPutData(data: shortstring);
begin
 writeStrToFile(fn,data);
end;

class procedure TEv3Data.WriteStrToFile(filename: shortstring; data: shortstring
  );
 var ff:cint;
begin
 ff:=fpopen(filename,O_WrOnly);
 fpseterrno(0);
 data+=#10;
 fpwrite(ff,data[1],length(data));
 fpseterrno(0);
 fpclose(ff);
 fpseterrno(0);
end;

class function TEv3Data.ReadStrFromFile(filename: shortstring): shortstring;
 var ff:cint;
begin
 result:='';
 ff:=fpopen(filename,O_RdOnly);
 fpseterrno(0);
 result[0]:=chr(fppread(ff,result[1],255,0));
 dec(result[0]);
 fpseterrno(0);
 fpclose(ff);
 fpseterrno(0);
end;

end.

